import textwrap

import yaml
import yamllint.config  # type: ignore
from yamllint import linter  # type: ignore
from yamllint.rules import key_duplicates  # type: ignore

import pytest
import unittest.mock

import tests.utils as utils

from wikitools import article_parser
from wikitools import yaml_rules
from wikitools_cli.commands import check_yaml

@pytest.fixture
def linter_config():
    cfg = yamllint.config.YamlLintConfig(check_yaml.DEFAULT_CONFIG_CONTENT)
    check_yaml.install_custom_checks(cfg)
    yield cfg


@pytest.fixture
def front_matter():
    def front_matter_maker(*dicts):
        dumps = "".join(
            yaml.dump(d, default_flow_style=False, sort_keys=True, indent=2, Dumper=article_parser.Dumper)
            for d in dicts
        )
        return "---\n{}---\n".format(dumps)

    yield front_matter_maker


class TestYamlRules:
    def test__duplicate_keys(self, linter_config, front_matter):
        fm = front_matter(
            dict(outdated=True, tags=[1, 2, 3]),
            dict(outdated=True, outdated_translation=True)
        )
        issue = next(linter.run(fm, linter_config))
        assert issue.rule == key_duplicates.ID

    def test__unknown_tags(self, linter_config, front_matter):
        fm = front_matter(
            dict(outdate=True, needs_cleanup=True, taggs=[1, 2])
        )
        first_issue, second_issue = list(linter.run(fm, linter_config))

        assert first_issue.rule == yaml_rules.AllowedTagsRule.ID
        assert first_issue.line == 3 and first_issue.column == 1  # "---" -> "needs_cleanup: true" -> "outdate: true"
        assert first_issue.level == 'error'
        assert "'outdate'" in first_issue.message

        assert second_issue.rule == yaml_rules.AllowedTagsRule.ID
        assert second_issue.line == 4 and second_issue.column == 1  # "---" -> "needs_cleanup: true" -> "outdate: true"
        assert second_issue.level == 'error'
        assert "'taggs'" in second_issue.message

    @pytest.mark.parametrize(
        "payload",
        [
            {"outdated": {"outdated": True}},
            {"tags": [["some-tag", "some-tag-2"], "some-tag-3"]},
            {"tags": [{"outdated": True}, "some-tag"]}
        ]
    )
    def test__bad_nesting(self, linter_config, front_matter, payload):
        fm = front_matter(payload)
        issue = next(linter.run(fm, linter_config))

        assert issue.rule == yaml_rules.NestedStructureRule.ID
        assert issue.line == 3  # points at the first nested element
        assert issue.level == 'error'

    def test__good_nesting(self, linter_config, front_matter, mocker: unittest.mock.Mock):
        rule = linter_config.rules[yaml_rules.NestedStructureRule.ID]
        rule.inner_check = mocker.Mock(side_effect=rule.inner_check)

        fm = front_matter(dict(tags=[1, 2, 3]))
        with pytest.raises(StopIteration):
            _ = next(linter.run(fm, linter_config))

        assert rule.inner_check.called

    def test__bad_top_level(self, linter_config, front_matter):
        fm = front_matter([
            "outdated",
            True,
        ])
        issue = next(linter.run(fm, linter_config))

        assert issue.rule == yaml_rules.TopLevelRule.ID
        assert issue.line == 2
        assert issue.level == 'error'

    def test__cli__read_front_matter(self, linter_config, root):
        utils.create_files(
            root,
            (
                'wiki/Article/en.md',
                textwrap.dedent('''
                    ---
                    stub: true
                    unknown_tag: true
                    tags:
                        - outdated: true
                    ---

                    # An article
                ''').strip()
            )
        )

        fm = check_yaml.read_yaml(str(root / 'wiki/Article/en.md'))
        tag_issue, nesting_issue = list(linter.run(fm, linter_config))

        assert tag_issue.rule == yaml_rules.AllowedTagsRule.ID
        assert tag_issue.line == 3

        assert nesting_issue.rule == yaml_rules.NestedStructureRule.ID
        assert nesting_issue.line == 5



import collections
import textwrap
from urllib import parse

import tests.conftest
import tests.utils as utils

from wikitools import article_parser, reference_parser


class TestArticleParser:
    def test__read_article(self, root):
        utils.create_files(
            root,
            (
                'wiki/Article/en.md',
                textwrap.dedent('''
                    ---
                    stub: true
                    tags:
                      - k1
                      - m1
                    ---

                    # An article

                    Links! [Links](https://example.com)!

                    Links, [zwo](/wiki/Article_two), [drei](Nested_article), [vier][vier_ref]!

                    [Links][links_ref]!

                    [links_ref]: https://example.com

                    ## List of references

                    [vier_ref]: /wiki/Article_three "Links!"
                ''').strip()
            )
        )

        article = article_parser.parse('wiki/Article/en.md')

        assert article.directory == 'wiki/Article'
        assert article.filename == 'en.md'
        assert article.path == 'wiki/Article/en.md'
        assert article.identifiers == {'list-of-references': 18}
        assert article.references == {
            'links_ref': reference_parser.Reference(
                lineno=16, name='links_ref', raw_location='https://example.com',
                parsed_location=parse.urlparse('https://example.com'), title=''
            ),
            'vier_ref': reference_parser.Reference(
                lineno=20, name='vier_ref', raw_location='/wiki/Article_three',
                parsed_location=parse.urlparse('/wiki/Article_three'), title='Links!'
            )
        }
        assert article.front_matter["stub"] is True
        assert article.front_matter["tags"] == ["k1", "m1"]

        assert set(article.lines.keys()) == {10, 12, 14}
        # lines are stored as-is, with trailing line breaks
        assert article.lines[10].raw_line == 'Links! [Links](https://example.com)!\n'

    def test__read_newspost(self, root):
        utils.create_files(
            root,
            (
                'news/newspost.md',
                textwrap.dedent('''
                    ---
                    layout: post
                    title: News!!!
                    date: 2021-10-21 15:00:00 +0000
                    ---

                    Links! [Links](https://example.com)!

                    Links, [zwo](/wiki/Article_two), [drei](Nested_article), [vier][vier_ref]!

                    [Links][links_ref]!

                    [links_ref]: https://example.com

                    ## List of references

                    [vier_ref]: /wiki/Article_three "Links!"
                ''').strip()
            )
        )

        article = article_parser.parse('news/newspost.md')

        assert article.directory == 'news'
        assert article.filename == 'newspost.md'
        assert article.path == 'news/newspost.md'
        assert article.identifiers == {'list-of-references': 15}
        assert article.references == {
            'links_ref': reference_parser.Reference(
                lineno=13, name='links_ref', raw_location='https://example.com',
                parsed_location=parse.urlparse('https://example.com'), title=''
            ),
            'vier_ref': reference_parser.Reference(
                lineno=17, name='vier_ref', raw_location='/wiki/Article_three',
                parsed_location=parse.urlparse('/wiki/Article_three'), title='Links!'
            )
        }
        assert article.front_matter["layout"] == "post"
        assert article.front_matter["title"] == "News!!!"
        assert article.front_matter["date"] == "2021-10-21 15:00:00 +0000"

        assert set(article.lines.keys()) == {7, 9, 11}
        # lines are stored as-is, with trailing line breaks
        assert article.lines[7].raw_line == 'Links! [Links](https://example.com)!\n'

    def test__read_article__with_comments(self, root):
        utils.create_files(
            root,
            (
                'wiki/Article/en.md',
                textwrap.dedent('''
                    # An article

                    <!-- rewrite this? do we even need it?
                        Hear the poetry:
                    -->

                    Roses are [red](/wiki/Red),
                    Violets are [blue][blue_ref] ![](img/violet.png),
                    I've written a program
                    <!-- Which didn't have [a clue](/wiki/Clue) -->
                    But neither should you.

                    <!--
                    Multiline [comments](/wiki/Comment)?
                    In my [test](/wiki/Not_a_test)?

                    [blue_ref]: /wiki/Blue
                    A wild {id=identifier}
                    -->

                    <!-- Another wild {#identifier} -->
                ''').strip()
            )
        )

        article = article_parser.parse('wiki/Article/en.md')
        assert len(article.references) == 0  # commented references are also skipped
        assert not article.front_matter

        links = sum((line.links for line in article.lines.values()), start=[])
        locations = set(_.raw_location for _ in links)
        assert locations == {'/wiki/Red', 'img/violet.png', 'blue_ref'}
        assert article.identifiers == {}

    def test__repeating_headings(self, root):
        utils.create_files(
            root,
            (
                'wiki/Ranking_criteria/en.md',
                textwrap.dedent('''
                    # Ranking criteria

                    ## Section

                    ## Section

                    ## Something else

                    ## Random
                    
                    <!-- A {#random} comment -->

                    ## Tricky section {#random}

                    ## Section
                ''').strip()
            )
        )

        article = article_parser.parse('wiki/Ranking_criteria/en.md')
        assert article.identifiers == {
            'section': 3,
            'section.1': 5,
            'something-else': 7,
            'random': 9,
            'random.1': 13,
            'section.2': 15,
        }

    def test__ignore_comments(self, root):
        utils.create_files(
            root,
            (
                'wiki/Comments/en.md',
                textwrap.dedent('''
                    # Comments

                    <!-- Don't mention [comments](/wiki/HTML#comment). -->

                    <!-- Don't mention the [comments](/wiki/HTML#comment) at all.
                        Yes, even if they span across several [lines](/wiki/Power_line).
                        Please be [silent](/wiki/Silence) about that, okay? --> [Test](/wiki/Test)

                    There is [no](/wiki/No) support<!-- for the [comments](/wiki/HTML#comment) --> on the wiki.
                ''').strip()
            )
        )
        article = article_parser.parse('wiki/Comments/en.md')
        assert set(article.lines.keys()) == {7, 9}

        assert len(article.lines[7].links) == 1
        assert article.lines[7].links[0].raw_location == "/wiki/Test"

        assert len(article.lines[9].links) == 1
        assert article.lines[9].links[0].raw_location == "/wiki/No"

    def test__ignore_code_blocks(self, root):
        utils.create_files(
            root,
            (
                'wiki/Code_blocks/en.md',
                textwrap.dedent('''
                    # Code blocks

                    ## Examples

                    `[Inline](/wiki/Inline)` | `[b][i]Inline[/i][/b]`

                    `` `[Also inline](/wiki/Also_inline)` ``

                    Let's take a [break](/wiki/Gameplay/Break)!

                    ``Some`` [fun stuff](/wiki/Fun_stuff) ``here``.

                    ```
                    [Multiline](/wiki/Multiline)
                    [b][i]No[/i][/b]
                    ```

                    ```markdown
                    [Multiline with syntax highlighting](/wiki/Multiline#syntax-highlighting)
                    [wow][wow_ref]

                    [wow_ref]: /wiki/Wow
                    ```
                ''').strip()
            )
        )
        article = article_parser.parse('wiki/Code_blocks/en.md')
        assert set(article.lines.keys()) == {9, 11}

        assert len(article.lines[9].links) == 1
        assert article.lines[9].links[0].raw_location == "/wiki/Gameplay/Break"

        assert len(article.lines[11].links) == 1
        assert article.lines[11].links[0].raw_location == "/wiki/Fun_stuff"


class TestFrontMatter:
    def test__read_write_to_existing_front_matter(self, root):
        cases = ["", "<!-- a comment -->\n\n", "<div> some html </div>\n\n"]
        for test_case in cases:
            article_path = root.join("en.md")
            article_path.write_text(textwrap.dedent('''
                ---
                tags:
                  - a
                  - aaa
                  - юниcode
                outdated: true
                ---

                {}# Test

                Lorem (ipsum).
            ''').format(test_case).strip(), encoding='utf-8')

            with article_path.open("r", encoding='utf-8') as fd:
                fm = article_parser.load_front_matter(fd)

            assert collections.OrderedDict(fm) == collections.OrderedDict({
                'tags': ['a', 'aaa', 'юниcode'],
                'outdated': True,
            })

            fm['outdated_since'] = '0000b4dc0ffee000'
            article_parser.save_front_matter(str(article_path), fm)

            new_contents = article_path.read_text(encoding='utf-8')
            assert new_contents == textwrap.dedent('''
                ---
                tags:
                  - a
                  - aaa
                  - юниcode
                outdated: true
                outdated_since: 0000b4dc0ffee000
                ---

                {}# Test

                Lorem (ipsum).
            ''').format(test_case).strip()

    def test__read_write_to_no_existing_front_matter(self, root):
        cases = ["", "<!-- a comment -->\n\n", "<div> some html </div>\n\n"]
        for test_case in cases:
            article_path = root.join("en.md")
            article_path.write_text(textwrap.dedent('''
                {}# Test

                Lorem (ipsum).
            ''').format(test_case).strip(), encoding='utf-8')

            with open(article_path, "r", encoding='utf-8') as fd:
                front_matter = article_parser.load_front_matter(fd)
            front_matter['tags'] = ['a', 'aaa', 'юниcode']
            front_matter['outdated'] = True
            article_parser.save_front_matter(article_path, front_matter)

            with article_path.open("r", encoding='utf-8') as fd:
                print("Article contents:\n" + fd.read())
                fd.seek(0)
                fm = article_parser.load_front_matter(fd)

            assert collections.OrderedDict(fm) == collections.OrderedDict({
                'tags': ['a', 'aaa', 'юниcode'],
                'outdated': True,
            })

            fm['outdated_since'] = '0000b4dc0ffee000'
            article_parser.save_front_matter(str(article_path), fm)

            new_contents = article_path.read_text(encoding='utf-8')
            assert new_contents == textwrap.dedent('''
                ---
                tags:
                  - a
                  - aaa
                  - юниcode
                outdated: true
                outdated_since: 0000b4dc0ffee000
                ---

                {}# Test

                Lorem (ipsum).
            ''').format(test_case).strip()



from wikitools import identifier_parser


class TestIdentifierParser:
    def test__plain_headings(self):
        for heading, identifier in (
            ('# Game modifiers', None),
            ('## Game modifiers', 'game-modifiers'),
            ('### Game modifiers', 'game-modifiers'),
            ('#### Game modifiers', 'game-modifiers'),
        ):
            assert identifier_parser.extract_identifier(heading) == (identifier, 0)

    # this uses real-life examples
    def test__punctuation(self):
        for heading, identifier in (
            (
                "## I've forgotten my username and password!",
                "i've-forgotten-my-username-and-password!"
            ), (
                "## What is this 'Bancho authentication error' I keep receiving?",
                "what-is-this-'bancho-authentication-error'-i-keep-receiving?"
            ), (
                '## What is "restricted" mode, exactly?',
                'what-is-"restricted"-mode,-exactly?'
            ), (
                '### Can someone make this skin from that show/game?',
                'can-someone-make-this-skin-from-that-show/game?'
            ),
        ):
            assert identifier_parser.extract_identifier(heading) == (identifier, 0)

    # this uses real-life examples
    def test__escape_sequences(self):
        for heading, identifier in (
            (
                r"## \[Colours\]",
                "[colours]"
            ), (
                r"## Step \#1",
                "step-#1"
            ), (
                r"#### Чи я можу грати на тому ком\'ютері, який osu! користувач раніше використовував?",
                "чи-я-можу-грати-на-тому-ком'ютері,-який-osu!-користувач-раніше-використовував?"
            ), ( # except these ones
                r"#### A \ B",
                r"a-\-b"
            ), (
                r"#### A \\ B",
                r"a-\-b"
            ), (
                r"#### A \\\ B",
                r"a-\\-b"
            ), (
                r"#### A \\\\ B",
                r"a-\\-b"
            ),
        ):
            assert identifier_parser.extract_identifier(heading) == (identifier, 0)

    def test__figure(self):
        for heading, identifier in (
            ('### ![osu! icon](/wiki/shared/mode/osu.png) pippi', 'pippi'),
            ('### Mani ![osu!mania icon](/wiki/shared/mode/mania.png) Mari', 'mani-mari'),
            ('### osu! ![][osu!]', 'osu!'),
        ):
            assert identifier_parser.extract_identifier(heading) == (identifier, 0)

    def test__link(self):
        for heading, identifier in (
            ('## [accounts@example.com](mailto:accounts@example.com)', 'accounts@example.com'),
            ('## I dare you, I [double dare you](/wiki/Say_what_again)', 'i-dare-you,-i-double-dare-you'),
            ('## A [b](/wiki/B) c d!', 'a-b-c-d!'),
            ('## A [wild](/wiki/B) l[ink](/wiki/Ink) appears ![abc](/img/abc.png)', 'a-wild-link-appears'),
        ):
            assert identifier_parser.extract_identifier(heading) == (identifier, 0)

    def test__custom(self):
        for line, identifier, pos in (
            ('osu! is a free-to-win game.', None, 0),
            ('## How to play better {#get-good}', 'get-good', 24),
            ('## osu.ppy.sh {id=website}', 'website', 18),
            ('A regular line, but with an anchor. {id=tag}', 'tag', 40),
            ('{id=only-identifier-here}', 'only-identifier-here', 4),
            ('Now this is a story all about how my life got flipped. {#turned-upside-down}', 'turned-upside-down', 57),
        ):
            assert identifier_parser.extract_identifier(line) == (identifier, pos)

    def test__unicode(self):
        for heading, identifier in (
            ('### Что случится, если я нарушу правила?', 'что-случится,-если-я-нарушу-правила?'),
            ('### 当我违反规定时会发生什么？', '当我违反规定时会发生什么？'),
            ('### Écran des résultats', 'écran-des-résultats'),
        ):
            assert identifier_parser.extract_identifier(heading) == (identifier, 0)



import typing

Redirects = typing.Dict[str, typing.Tuple[str, int]]


def unquote_and_trim(s):
    s = s.strip()
    if (s[0] == '"' and s[-1] == '"') or (s[0] == "'" and s[-1] == "'"):
        return s[1:-1]
    return s


def load_redirects(path: str) -> Redirects:
    """
    Read redirects from a YAML dictionary file (done manually to preserve line numbers).
    """

    redirects = {}
    with open(path, 'r', encoding='utf-8') as fd:
        for line_number, line in enumerate(fd, start=1):
            split = line.split(':')
            try:
                redirects[unquote_and_trim(split[0])] = (unquote_and_trim(split[1]), line_number)
            except IndexError:
                pass
    return redirects



import textwrap

from wikitools import redirect_parser


class TestRedirectParser:
    def test__parse_redirects(self, tmpdir):
        path = tmpdir.join('redirect.yaml')
        path.write(textwrap.dedent('''
            # some junk on top

            "asc": "Article_Styling_Criteria"
            "asc/images": "Article_Styling_Criteria#images"

            "ignore_list": "Client/Options/Ignore_list"
            "ignore":      "Client/Options/Ignore_list"
            unquoted_key1:  unquoted/value1
            "quoted_key": unquoted/value2
            unquoted_key2:    "quoted/value"
        ''').strip())

        redirects = redirect_parser.load_redirects(str(path))
        assert redirects == {
            'asc': ('Article_Styling_Criteria', 3),
            'asc/images': ('Article_Styling_Criteria#images', 4),
            'ignore_list': ('Client/Options/Ignore_list', 6),
            'ignore': ('Client/Options/Ignore_list', 7),
            'unquoted_key1':  ('unquoted/value1', 8),
            'quoted_key':  ('unquoted/value2', 9),
            'unquoted_key2':  ('quoted/value', 10),
        }



import fnmatch
import itertools
import os
import typing
import pathlib


class ChangeDirectory:
    cwd: str

    def __init__(self, repo_root: str):
        self.cwd = os.getcwd()
        os.chdir(repo_root)

    def __del__(self):
        os.chdir(self.cwd)


def normalised(path: str) -> str:
    normalised = os.path.normpath(path).replace("\\", "/")
    if normalised.startswith("./"):
        normalised = normalised[2:]
    return normalised


"""
Returns a dictionary of file and directory paths, with a lowercased path for the key and the original casing for the value.
Useful for looking up file paths case-insensitively on case-sensitive file systems.
"""
def file_tree():
    # this cache would only become invalid when the current working directory changes, which only happens in tests and not during normal execution
    if not hasattr(file_tree, "cache"):
        tree = {normalised(article_path.lower()): normalised(article_path) for article_path in itertools.chain(list_all_dirs(["."]), list_all_files(["."]))}
        setattr(file_tree, "cache", tree)
    return getattr(file_tree, "cache")


def get_canonical_path_casing(path: pathlib.Path) -> pathlib.Path:
    """
    Converts a file/directory path into a path with the correct casing.
    The path must exist (throws KeyError otherwise)
    """

    return pathlib.Path(file_tree()[normalised(os.path.relpath(path.as_posix()).lower())])


def exists_case_sensitive(path: pathlib.Path) -> bool:
    """
    Case-sensitive file/directory existence check
    """

    if os.name == 'nt':
        try:
            path_string = path.as_posix()
            # windows disallows two files that differ only in casing, so there are no special considerations for that
            # os.path.realpath with strict=True does the existence check
            return normalised(path_string) == normalised(os.path.relpath(os.path.realpath(path_string, strict=True)))
        except OSError:
            return False
    else:
        return path.exists()


def exists_case_insensitive(path: pathlib.Path) -> bool:
    """
    Case-insensitive file/diretory existence check
    """

    if os.name == 'nt':
        return path.exists()
    else:
        # case-insensitive directory/file existence checking isn't trivial in case-sensitive file systems because os-provided existence checks can't be relied upon

        return normalised(os.path.relpath(path.as_posix()).lower()) in file_tree()


def is_newspost(path: str) -> bool:
    return (
        normalised(os.path.dirname(path)).startswith("news") and
        path.endswith(".md")
    )


def is_article(path: str) -> bool:
    filename = os.path.basename(path)
    return (
        fnmatch.fnmatch(filename, "??.md") or
        fnmatch.fnmatch(filename, "???.md") or
        fnmatch.fnmatch(filename, "??-??.md")
    )


def is_translation(path: str) -> bool:
    filename = os.path.basename(path)
    return (
        filename != "en.md" and
        (
            fnmatch.fnmatch(filename, "??.md") or
            fnmatch.fnmatch(filename, "???.md") or
            fnmatch.fnmatch(filename, "??-??.md")
        )
    )


def is_original(path: str) -> bool:
    return os.path.basename(path) == "en.md"


def list_all_files(roots: typing.Iterable[str]=["wiki"]) -> typing.Generator[str, None, None]:
    for item in roots:
        for root, _, filenames in os.walk(item):
            for f in filenames:
                filepath = os.path.join(root, f).replace("\\", "/")
                yield filepath


def list_all_dirs(roots: typing.Iterable[str]=["wiki"]) -> typing.Generator[str, None, None]:
    """
    List ALL directories
    """

    for item in roots:
        for root, _, __ in os.walk(item):
            yield root.replace("\\", "/")


def list_all_article_dirs() -> typing.Generator[str, None, None]:
    """
    List ALL article directories in the wiki
    """

    for root, _, filenames in os.walk("wiki"):
        if any(is_article(f) for f in filenames):
            yield root.replace("\\", "/")


def list_all_articles() -> typing.Generator[str, None, None]:
    """
    List ALL article files in the wiki
    """

    for filepath in list_all_files(["wiki"]):
        if is_article(filepath):
            yield filepath.replace("\\", "/")


def list_all_newsposts() -> typing.Generator[str, None, None]:
    """
    List ALL newsposts
    """

    for filepath in list_all_files(["news"]):
        if is_newspost(filepath):
            yield filepath.replace("\\", "/")


def list_all_articles_and_newsposts() -> typing.Generator[str, None, None]:
    """
    List ALL article and newspost files 
    """

    for filepath in list_all_files(["wiki", "news"]):
        if is_article(filepath) or is_newspost(filepath):
            yield filepath.replace("\\", "/")


def list_all_translations(article_dirs: typing.Iterable[str]) -> typing.Generator[str, None, None]:
    """
    List ALL translations inside the articles specified
    """

    for d in article_dirs:
        for filename in sorted(os.listdir(d)):
            if is_original(filename) or not is_article(filename):
                continue

            yield os.path.join(d, filename).replace("\\", "/")



from tests.conftest import get_visual_tests
from tests.conftest import run_visual_test


tests = get_visual_tests()


class TestVisualTests:
    pass


def make_method(test_index, case_index):
    def method(self):
        global tests
        nonlocal test_index
        nonlocal case_index
        run_visual_test(tests, test_index, case_index)
    return method


for test in tests:
    for test_index, test in enumerate(tests):
        test_slug = test.name.lower().replace(" ", "_")
        for case_index, test_case in enumerate(test.cases):
            setattr(TestVisualTests, f"test_visual__{test_slug}__{test_case.name}", make_method(test_index, case_index))



import textwrap
from urllib import parse

from wikitools import reference_parser


class TestReferences:
    def test__no_referenes(self):
        assert reference_parser.extract('A totally normal line.', lineno=1) is None
        assert reference_parser.extract('A line with [a link](/wiki/Link).', lineno=1) is None

    def test__references(self):
        assert reference_parser.extract('[refname]: /wiki/Article', lineno=10) == reference_parser.Reference(
            lineno=10,
            name='refname',
            raw_location='/wiki/Article', parsed_location=parse.urlparse('/wiki/Article'),
            title='',
        )

        assert reference_parser.extract('[ref]: /some/path "Title"', lineno=11) == reference_parser.Reference(
            lineno=11,
            name='ref',
            raw_location='/some/path', parsed_location=parse.urlparse('/some/path'),
            title='Title'
        )

        assert reference_parser.extract(
            '[ref]: https://example.com/image.png "Image"', lineno=12
        ) == reference_parser.Reference(
            lineno=12,
            name='ref',
            raw_location='https://example.com/image.png',
            parsed_location=parse.urlparse('https://example.com/image.png'),
            title='Image'
        )


class TestReferenceFinder:
    def test__extract_all(self):
        text = textwrap.dedent('''
            # An article

            [stray]: /refe/ren/ce
        ''').strip()

        expected_reference = reference_parser.Reference(
            lineno=3, name='stray', raw_location='/refe/ren/ce',
            parsed_location=parse.urlparse('/refe/ren/ce'), title=''
        )
        assert reference_parser.extract_all(text) == {'stray': expected_reference}

    def test__more_references(self):
        text = textwrap.dedent('''
            # Lorem ipsum

            Dolor [sit][sit_ref] amet.

            [sit_ref]: /a/random/insertion

            It is a long established fact that a ![reader][reader_ref] will be distracted by... [KEEP READING]

            [reader_ref]: img/reader.png "A reader"
        ''').strip()

        link_ref = reference_parser.Reference(
            lineno=5, name='sit_ref', raw_location='/a/random/insertion',
            parsed_location=parse.urlparse('/a/random/insertion'), title=''
        )
        image_ref = reference_parser.Reference(
            lineno=9, name='reader_ref', raw_location='img/reader.png',
            parsed_location=parse.urlparse('img/reader.png'), title='A reader'
        )

        assert reference_parser.extract_all(text) == {
            'sit_ref': link_ref,
            'reader_ref': image_ref
        }



import tests.conftest
import tests.utils as utils

from wikitools_cli.commands import check_links as link_checker


class TestCheckLinks:
    def test__check_links_all_valid(self, root):
        article_paths = [
            'wiki/redirect.yaml',
            'wiki/Article/en.md',
            'wiki/Article/pt-br.md',
            'wiki/Article/zh-tw.md',
            'wiki/Category1/Article/en.md',
            'wiki/Category1/Article/fr.md',
            'wiki/Category1/Article/zh-tw.md',
            'wiki/Category1/Article/TEMPLATE.md',
            'news/2023/newspost.md',
        ]

        utils.create_files(root, *((path, '[good link](/wiki/Article)') for path in article_paths))

        exit_code = link_checker.main("--all")
        assert exit_code == 0

    def test__check_links_all_invalid(self, root):
        article_paths = [
            'wiki/redirect.yaml',
            'wiki/Article/en.md',
            'wiki/Article/pt-br.md',
            'wiki/Article/zh-tw.md',
            'wiki/Category1/Article/en.md',
            'wiki/Category1/Article/fr.md',
            'wiki/Category1/Article/zh-tw.md',
            'wiki/Category1/Article/TEMPLATE.md',
            'news/2023/newspost.md',
        ]

        utils.create_files(root, *((path, '[bad link](/wiki/Not_an_article)') for path in article_paths))

        exit_code = link_checker.main("--all")
        assert exit_code == 1

    def test__check_links_all_valid_with_changed_root(self, root):
        article_paths = [
            'root/wiki/redirect.yaml',
            'root/wiki/Article/en.md',
            'root/wiki/Article/pt-br.md',
            'root/wiki/Article/zh-tw.md',
            'root/wiki/Category1/Article/en.md',
            'root/wiki/Category1/Article/fr.md',
            'root/wiki/Category1/Article/zh-tw.md',
            'root/wiki/Category1/Article/TEMPLATE.md',
            'root/news/2023/newspost.md',
        ]

        utils.create_files(root, *((path, '[good link](/wiki/Article)') for path in article_paths))

        exit_code = link_checker.main("--all", "--root", "root")
        assert exit_code == 0

    def test__check_links_all_invalid_with_changed_root(self, root):
        article_paths = [
            'root/wiki/redirect.yaml',
            'root/wiki/Article/en.md',
            'root/wiki/Article/pt-br.md',
            'root/wiki/Article/zh-tw.md',
            'root/wiki/Category1/Article/en.md',
            'root/wiki/Category1/Article/fr.md',
            'root/wiki/Category1/Article/zh-tw.md',
            'root/wiki/Category1/Article/TEMPLATE.md',
            'root/news/2023/newspost.md',
        ]

        utils.create_files(root, *((path, '[bad link](/wiki/Not_an_article)') for path in article_paths))

        exit_code = link_checker.main("--all", "--root", "root")
        assert exit_code == 1

    def test__check_specific_target_all_invalid(self, root):
        article_paths = [
            'wiki/redirect.yaml',
            'wiki/Article/en.md',
            'news/2023/newspost.md',
        ]

        utils.create_files(root, *((path, '[bad link](/wiki/Not_an_article)') for path in article_paths))

        for article in article_paths[1:]:
            print(article)
            exit_code = link_checker.main("--target", article)
            assert exit_code == 1

    def test__check_specific_target_all_valid(self, root):
        article_paths = [
            'wiki/redirect.yaml',
            'wiki/Article/en.md',
            'news/2023/newspost.md',
        ]

        utils.create_files(root, *((path, '[good link](/wiki/Article)') for path in article_paths))

        for article in article_paths[1:]:
            print(article)
            exit_code = link_checker.main("--target", article)
            assert exit_code == 0



import os
import pathlib
import typing
import urllib

from wikitools import redirect_parser, reference_parser, errors, link_parser, article_parser
from wikitools import console
from wikitools.file_utils import exists_case_sensitive, exists_case_insensitive
from wikitools.file_utils import get_canonical_path_casing
from wikitools.file_utils import is_article


class PathType:
    """
    Additional information about an article path

    Wiki articles support redirects, while news posts don't
    News post also don't have translations
    """

    WIKI = 0
    NEWS = 1
    GITHUB = 2


class RepositoryPath(typing.NamedTuple):
    """
    Relative path from the ppy/osu-wiki root to a news post markdown file or wiki article directory
    """

    path_type: int
    path: pathlib.Path
    fragment: typing.Optional[str]


def is_fragment_only(parsed_location: urllib.parse.ParseResult):
    return parsed_location.fragment and not any((parsed_location.scheme,
        parsed_location.netloc, parsed_location.path,
        parsed_location.params, parsed_location.query))


def is_news_link(parsed_location: urllib.parse.ParseResult):
    return ((parsed_location.scheme == "http" or parsed_location.scheme == "https") and
        parsed_location.netloc == "osu.ppy.sh" and parsed_location.path.startswith("/home/news/"))


def is_github_link(parsed_location: urllib.parse.ParseResult):
    return (
        (parsed_location.scheme == "http" or parsed_location.scheme == "https")
        and parsed_location.netloc == "github.com"
        and (parsed_location.path.startswith("/ppy/osu-wiki/blob/master/") or parsed_location.path.startswith("/ppy/osu-wiki/tree/master/"))
    )


def get_repo_path(
    current_article: pathlib.Path,
    link: link_parser.Link,
    parsed_location: urllib.parse.ParseResult
) -> typing.Union[RepositoryPath, errors.LinkError, None]:
    """
    Converts a wild link of into a osu-wiki repository path with an optional fragment, if possible

    Acceptable links can be in the following formats:
    
    - Relative wiki link: path/to/article#optional-fragment
    - Absolute wiki link: /wiki/path/to/article#optional-fragment
    - News post link: https://osu.ppy.sh/home/news/2023-01-01-news#optional-fragment
    - Section link within current article or news post: #fragment
    """

    if is_fragment_only(parsed_location):
        path_type = PathType.NEWS if current_article.as_posix().startswith("news") else PathType.WIKI
        return RepositoryPath(path_type=path_type, path=current_article, fragment=parsed_location.fragment)

    if is_news_link(parsed_location):
        file = pathlib.Path(parsed_location.path[1:] + ".md").relative_to("home").name
        year = file.split("-")[0]
        path = pathlib.Path(f"news/{year}/{file}")
        return RepositoryPath(path_type=PathType.NEWS, path=path, fragment=parsed_location.fragment)

    if is_github_link(parsed_location):
        path = pathlib.Path("/".join(parsed_location.path.split("/")[5:]))
        return RepositoryPath(path_type=PathType.GITHUB, path=path, fragment=parsed_location.fragment)

    # some external link; don't care
    if parsed_location.scheme:
        return None

    if is_article(parsed_location.path):
        return errors.MalformedLinkError(link, "wiki links must not include the article file name")

    if parsed_location.path.startswith("/"):
        # absolute wiki link
        path = pathlib.Path(parsed_location.path[1:])
        return RepositoryPath(path_type=PathType.WIKI, path=path, fragment=parsed_location.fragment)

    # domain is non-empty, but the link is internal?
    if parsed_location.netloc:
        return errors.MalformedLinkError(link, "incorrect link structure (typo?)")

    # relative wiki link
    path = current_article / parsed_location.path
    return RepositoryPath(path_type=PathType.WIKI, path=path, fragment=parsed_location.fragment)


def resolve_redirect(
    repo_path: RepositoryPath,
    link: link_parser.Link,
    reference: typing.Optional[reference_parser.Reference],
    redirects: redirect_parser.Redirects,
    exists: typing.Callable[[pathlib.Path], bool]
) -> typing.Union[RepositoryPath, errors.LinkError]:
    """
    Resolves a wiki article path according to redirects.

    Returns an error if the redirect or article does not exist
    """

    redirect_source = repo_path.path.relative_to("wiki").as_posix()
    try:
        redirect_destination, redirect_line_no = redirects[redirect_source.lower()]
    except KeyError:
        return errors.LinkNotFoundError(link, reference, repo_path.path.as_posix())

    split = redirect_destination.split('#')
    path = pathlib.Path('wiki') / split[0]
    fragment = split[1] if len(split) > 1 else None
    target_path = RepositoryPath(path_type=repo_path.path_type, path=path, fragment=fragment)

    if not exists(target_path.path):
        return errors.BrokenRedirectError(link, redirect_source, redirect_line_no, redirect_destination)

    return RepositoryPath(path_type=repo_path.path_type, path=path, fragment=fragment)


def check_link(
    article: article_parser.Article, link: link_parser.Link,
    redirects: redirect_parser.Redirects, references: reference_parser.References,
    all_articles: typing.Dict[str, article_parser.Article],
    case_sensitive: bool = False
) -> typing.Optional[errors.LinkError]:
    """
    Verifies that the link is valid:
        - External links are always assumed valid, since we can't just issue HTTP requests left and right
        - For Markdown references, there exists a dereferencing line with [reference_name]: /lo/ca/ti/on
        - Direct internal links, as well as redirects, must point to existing article files
        - Relative links are parsed under the assumption that they are located inside the current article's directory
    """

    if case_sensitive:
        exists = exists_case_sensitive
    else:
        exists = exists_case_insensitive

    reference = link.resolve(references)
    if reference is None and link.is_reference:
        return errors.MissingReferenceError(link)

    parsed_location = reference.parsed_location if reference else link.parsed_location

    current_article_path = pathlib.Path(article.path)
    if current_article_path.as_posix().startswith("wiki"):
        current_article_path = pathlib.Path(os.path.dirname(current_article_path))

    repo_path = get_repo_path(current_article_path, link, parsed_location)
    if isinstance(repo_path, errors.LinkError):
        return repo_path
    elif repo_path is None:
        return None

    # github gives a 404 when the casing is wrong
    if repo_path.path_type == PathType.GITHUB:
        exists = exists_case_sensitive

    if not exists(repo_path.path):
        # if the article doesn't exist, check if it has a redirect
        if repo_path.path_type == PathType.NEWS or repo_path.path_type == PathType.GITHUB:
            # except news and github links don't support redirects
            return errors.LinkNotFoundError(link, reference, repo_path.path.as_posix())

        redirected_path = resolve_redirect(repo_path, link, reference, redirects, exists)
        if isinstance(redirected_path, errors.LinkError):
            return redirected_path
        repo_path = redirected_path

    # link to an article in general, article exists -> good
    if not parsed_location.fragment:
        return None

    target_path = repo_path.path
    if os.name != 'nt' and not case_sensitive:
        target_path = get_canonical_path_casing(target_path)

    # link to a section
    match repo_path.path_type:
        case PathType.GITHUB:
            # github links can either be directories or files
            # but section links are only relevant for markdown files
            if repo_path.path.suffix == ".md":
                raw_path = target_path.as_posix()
                if raw_path not in all_articles:
                    # this is safe to do since the caller iterates over a copy of all_articles -> we can modify it as we wish
                    all_articles[raw_path] = article_parser.parse(target_path)

                target_article = all_articles[raw_path]

                if parsed_location.fragment not in target_article.identifiers:
                    # collect some additional metadata before reporting
                    translation_outdated = False
                    if repo_path.path.name != "en.md":
                        translation_outdated = target_article.front_matter.get('outdated_translation', False)

                    return errors.MissingIdentifierError(link, raw_path, parsed_location.fragment, False, translation_outdated)
        case PathType.NEWS:
            # always a file path
            raw_path = target_path.as_posix()
            if raw_path not in all_articles:
                all_articles[raw_path] = article_parser.parse(target_path)
            target_article = all_articles[raw_path]

            if parsed_location.fragment not in target_article.identifiers:
                return errors.MissingIdentifierError(link, raw_path, parsed_location.fragment, False, False)
        case PathType.WIKI:
            # directory -> need to find the target article; it could be a translation
            # XXX(TicClick): this part assumes there is always an English version of the article in a folder
            target_file = target_path / article.filename
            translation = target_file # verified to be the case later
            no_translation_available = article.filename != 'en.md' and not target_file.is_file()

            if no_translation_available:
                target_file = target_path / 'en.md'

            raw_path = target_file.as_posix()
            if raw_path not in all_articles:
                # this is safe to do since the caller iterates over a copy of all_articles -> we can modify it as we wish
                all_articles[raw_path] = article_parser.parse(target_file)

            target_article = all_articles[raw_path]

            if parsed_location.fragment not in target_article.identifiers:
                # collect some additional metadata before reporting
                translation_outdated = False
                if not no_translation_available:
                    raw_path_translation = translation.as_posix()
                    if raw_path_translation not in all_articles:
                        # this is safe to do since the caller iterates over a copy of all_articles -> we can modify it as we wish
                        all_articles[raw_path_translation] = article_parser.parse(translation)
                    translation_outdated = all_articles[raw_path_translation].front_matter.get('outdated_translation', False)

                return errors.MissingIdentifierError(link, raw_path, parsed_location.fragment, no_translation_available, translation_outdated)

    return None


def check_article(
    article: article_parser.Article, redirects: redirect_parser.Redirects,
    all_articles: typing.Dict[str, article_parser.Article],
    case_sensitive: bool = False
) -> typing.Dict[int, typing.List[errors.LinkError]]:
    """
    Try resolving links in the article to other articles or files.
    """

    errors = {}
    for lineno, line in article.lines.items():
        local_errors = [
            errors for errors in (
                check_link(article, link, redirects, article.references, all_articles, case_sensitive)
                for link in line.links
            )
            if errors
        ]
        if local_errors:
            errors[lineno] = local_errors

    return errors



from collections import Counter as multiset
import textwrap

import tests.conftest
import tests.utils as utils

from wikitools import article_parser, git_utils, file_utils

from wikitools_cli.commands import check_outdated_articles as outdater


class TestCheckOutdatedArticles:
    def test__list_modified_translations(self, root):
        utils.set_up_dummy_repo()
        article_paths = [
            'wiki/Article/en.md',
            'wiki/Article/fr.md',
            'wiki/Article/pt-br.md',
            'wiki/Article/zh-tw.md',
            'wiki/Category1/Article/en.md',
            'wiki/Category1/Article/fr.md',
            'wiki/Category1/Article/pt-br.md',
            'wiki/Category1/Article/zh-tw.md',
            'wiki/Category1/Article/TEMPLATE.md',
            'wiki/Category1/Category2/Article/en.md',
            'wiki/Category1/Category2/Article/fr.md',
            'wiki/Category1/Category2/Article/pt-br.md',
            'wiki/Category1/Category2/Article/zh-tw.md',
            'wiki/Category1/Category2/Category3/Article/en.md',
            'wiki/Category1/Category2/Category3/Article/fr.md',
            'wiki/Category1/Category2/Category3/Article/pt-br.md',
            'wiki/Category1/Category2/Category3/Article/zh-tw.md',
        ]

        utils.create_files(root, *((path, '') for path in article_paths))
        utils.stage_all_and_commit("initial commit")

        utils.create_files(root, *((path, '# Article') for path in article_paths))
        utils.stage_all_and_commit("add article title")
        commit_hash = utils.get_last_commit_hash()

        # note that at least two existing commits are necessary to get a diff using `revision^`
        modified_translations = outdater.list_modified_translations(commit_hash)

        assert multiset(modified_translations) == multiset(utils.remove(article_paths, "en.md", "TEMPLATE.md"))

        utils.create_files(root,
            *((path, '# Article\n\nCeci est un article en français.') for path in
            utils.take(article_paths, "fr.md"))
        )
        utils.stage_all_and_commit("add article content")
        commit_hash = utils.get_last_commit_hash()

        modified_translations = outdater.list_modified_translations(commit_hash)

        assert multiset(modified_translations) == multiset(utils.take(article_paths, "fr.md"))

    def test__list_modified_originals(self, root):
        utils.set_up_dummy_repo()
        article_paths = [
            'wiki/Article/en.md',
            'wiki/Article2/en.md',
            'wiki/Article/fr.md',
            'wiki/Article2/fr.md',
            'wiki/Article/pt-br.md',
            'wiki/Article2/pt-br.md',
            'wiki/Article/zh-tw.md',
            'wiki/Article2/zh-tw.md',
        ]

        utils.create_files(root, *((path, '# Article') for path in article_paths))
        utils.stage_all_and_commit("add some articles")
        commit_hash = utils.get_last_commit_hash()

        utils.create_files(root, *zip(article_paths[0:2], [
            '# Article\n\nThis is an article in English.',
            '# Article\n\nThis is another article in English.',
        ]))
        utils.stage_all_and_commit("add article content")
        commit_hash = utils.get_last_commit_hash()

        modified_originals = outdater.list_modified_originals(commit_hash)
        assert multiset(modified_originals) == multiset(utils.take(article_paths, "en.md"))

    def test__list_outdated_translations(self, root):
        utils.set_up_dummy_repo()
        article_paths = [
            'wiki/Article/en.md',
            'wiki/Article2/en.md',
            'wiki/Article/fr.md',
            'wiki/Article2/fr.md',
            'wiki/Article/pt-br.md',
            'wiki/Article2/pt-br.md',
            'wiki/Article/zh-tw.md',
            'wiki/Article2/zh-tw.md',
        ]

        utils.create_files(root, *((path, '# Article') for path in article_paths))
        utils.stage_all_and_commit("add some articles")

        utils.create_files(root, *zip(article_paths[0:2], [
            '# Article\n\nThis is an article in English.',
            '# Article\n\nThis is another article in English.',
        ]))
        utils.stage_all_and_commit("add english article content")
        translations_to_outdate = list(outdater.list_outdated_translations(
            set(utils.remove(article_paths, "en.md")),
            set()
        ))

        assert multiset(translations_to_outdate) == multiset(utils.remove(article_paths, "en.md"))

    def test__outdate_translations(self, root):
        utils.set_up_dummy_repo()
        article_paths = [
            'wiki/Article2/en.md',
            'wiki/Article/en.md',
            'wiki/Article2/fr.md',
            'wiki/Article/fr.md',
            'wiki/Article2/pt-br.md',
            'wiki/Article/pt-br.md',
            'wiki/Article2/zh-tw.md',
            'wiki/Article/zh-tw.md',
        ]

        utils.create_files(root, *((path, '# Article') for path in article_paths))
        utils.stage_all_and_commit("add some articles")

        utils.create_files(root, *zip(article_paths[0:2], [
            '# Article\n\nThis is an article in English.',
            '# Article\n\nThis is another article in English.',
        ]))
        utils.stage_all_and_commit("add english article content")
        commit_hash = utils.get_last_commit_hash()

        to_outdate_zh_tw = utils.take(article_paths, "zh-tw.md")
        outdater.outdate_translations(*to_outdate_zh_tw, outdated_hash=commit_hash)
        outdated_translations = utils.get_changed_files()
        utils.stage_all_and_commit("outdate zh-tw")

        assert multiset(outdated_translations) == multiset(to_outdate_zh_tw)

        to_outdate_all = utils.remove(article_paths, "en.md")
        outdater.outdate_translations(*to_outdate_all, outdated_hash=commit_hash)
        outdated_translations = utils.get_changed_files()
        utils.stage_all_and_commit("outdate the rest of the translations")

        assert multiset(outdated_translations) == multiset(utils.remove(article_paths, "en.md", "zh-tw.md"))

        for article in to_outdate_all:
            with open(article, "r", encoding='utf-8') as fd:
                content = fd.read()

            assert content == textwrap.dedent('''
                ---
                {}: true
                {}: {}
                ---

                # Article
            ''').strip().format(outdater.OUTDATED_TRANSLATION_TAG, outdater.OUTDATED_HASH_TAG, commit_hash)

    def test__validate_hashes(self, root):
        utils.set_up_dummy_repo()
        article_paths = [
            'wiki/Article/en.md',
            'wiki/Article/fr.md',
            'wiki/Article/pt-br.md',
            'wiki/Article/zh-tw.md',
        ]

        utils.create_files(root, *((path, '# Article') for path in article_paths))
        utils.stage_all_and_commit("add an article")

        utils.create_files(root, (article_paths[0], '# Article\n\nThis is an article in English.'))
        utils.stage_all_and_commit("modify english article")
        commit_hash = utils.get_last_commit_hash()

        outdater.outdate_translations(*article_paths[1:], outdated_hash=commit_hash)
        utils.stage_all_and_commit("outdate translations")

        with open(article_paths[1], "r", encoding='utf-8') as fd:
            front_matter = article_parser.load_front_matter(fd)
        front_matter[outdater.OUTDATED_HASH_TAG] = "bogus-commit-hash"
        article_parser.save_front_matter(article_paths[1], front_matter)
        utils.stage_all_and_commit("corrupt hash")

        assert multiset(outdater.check_commit_hashes(article_paths[1:])) == multiset(article_paths[1:2])

    def test__full_autofix_flow(self, root):
        utils.set_up_dummy_repo()
        article_paths = [
            'wiki/Article/en.md',
            'wiki/Article/fr.md',
            'wiki/Article/pt-br.md',
            'wiki/Article/zh-tw.md',
            'wiki/Category1/Article/en.md',
            'wiki/Category1/Article/fr.md',
            'wiki/Category1/Article/pt-br.md',
            'wiki/Category1/Article/zh-tw.md',
            'wiki/Category1/Category2/Article/en.md',
            'wiki/Category1/Category2/Article/fr.md',
            'wiki/Category1/Category2/Article/pt-br.md',
            'wiki/Category1/Category2/Article/zh-tw.md',
            'wiki/Category1/Category2/Category3/Article/en.md',
            'wiki/Category1/Category2/Category3/Article/fr.md',
            'wiki/Category1/Category2/Category3/Article/pt-br.md',
            'wiki/Category1/Category2/Category3/Article/zh-tw.md',
        ]

        utils.create_files(root, *((path, '# Article') for path in article_paths))
        utils.stage_all_and_commit("add articles")
        commit_hash_1 = utils.get_last_commit_hash()

        already_outdated_translations = utils.take(article_paths, "zh-tw.md")
        outdater.outdate_translations(*already_outdated_translations, outdated_hash=commit_hash_1)
        utils.stage_all_and_commit("outdate chinese translations")

        utils.create_files(root, *(
            (article_path, '# Article\n\nThis is an article in English.') for article_path in
            utils.take(article_paths, "en.md")
        ))
        utils.stage_all_and_commit("modify english articles")
        commit_hash_2 = utils.get_last_commit_hash()

        exit_code = outdater.main("--base-commit", commit_hash_2, f"{outdater.AUTOFIX_FLAG}")

        assert exit_code == 0

        outdated_translations = utils.get_changed_files()

        non_chinese_translations = utils.remove(article_paths, "en.md", "zh-tw.md")

        assert multiset(outdated_translations) == multiset(non_chinese_translations)

        expected_content = textwrap.dedent('''
            ---
            {}: true
            {}: {}
            ---

            # Article
        ''').strip()

        for article in already_outdated_translations:
            with open(article, "r", encoding='utf-8') as fd:
                content = fd.read()

            assert content == expected_content.format(outdater.OUTDATED_TRANSLATION_TAG, outdater.OUTDATED_HASH_TAG, commit_hash_1)

        for article in outdated_translations:
            with open(article, "r", encoding='utf-8') as fd:
                content = fd.read()

            assert content == expected_content.format(outdater.OUTDATED_TRANSLATION_TAG, outdater.OUTDATED_HASH_TAG, commit_hash_2)

        for article in utils.take(article_paths, "en.md"):
            with open(article, "r", encoding='utf-8') as fd:
                content = fd.read()

            assert content == '# Article\n\nThis is an article in English.'

        log = git_utils.git("--no-pager", "log", "--pretty=oneline").splitlines()

        assert len(log) == 3

    def test__full_autocommit_flow(self, root):
        utils.set_up_dummy_repo()
        article_paths = [
            'wiki/Article/en.md',
            'wiki/Article/fr.md',
            'wiki/Article/pt-br.md',
            'wiki/Article/zh-tw.md',
            'wiki/Category1/Article/en.md',
            'wiki/Category1/Article/fr.md',
            'wiki/Category1/Article/pt-br.md',
            'wiki/Category1/Article/zh-tw.md',
            'wiki/Category1/Category2/Article/en.md',
            'wiki/Category1/Category2/Article/fr.md',
            'wiki/Category1/Category2/Article/pt-br.md',
            'wiki/Category1/Category2/Article/zh-tw.md',
            'wiki/Category1/Category2/Category3/Article/en.md',
            'wiki/Category1/Category2/Category3/Article/fr.md',
            'wiki/Category1/Category2/Category3/Article/pt-br.md',
            'wiki/Category1/Category2/Category3/Article/zh-tw.md',
        ]

        utils.create_files(root, *((path, '# Article') for path in article_paths))
        utils.stage_all_and_commit("add articles")
        commit_hash_1 = utils.get_last_commit_hash()

        already_outdated_translations = utils.take(article_paths, "zh-tw.md")
        outdater.outdate_translations(*already_outdated_translations, outdated_hash=commit_hash_1)
        utils.stage_all_and_commit("outdate chinese translations")

        utils.create_files(root, *(
            (article_path, '# Article\n\nThis is an article in English.') for article_path in
            utils.take(article_paths, "en.md")
        ))
        utils.stage_all_and_commit("modify english articles")
        commit_hash_2 = utils.get_last_commit_hash()

        exit_code = outdater.main("--base-commit", commit_hash_2, f"{outdater.AUTOFIX_FLAG}", f"{outdater.AUTOCOMMIT_FLAG}")

        assert exit_code == 0

        commit_hash_3 = utils.get_last_commit_hash()

        assert commit_hash_3 != commit_hash_2

        assert utils.get_changed_files() == []

        outdated_translations = outdater.list_modified_translations(commit_hash_3)

        non_chinese_translations = utils.remove(article_paths, "en.md", "zh-tw.md")

        assert multiset(outdated_translations) == multiset(non_chinese_translations)

        expected_content = textwrap.dedent('''
            ---
            {}: true
            {}: {}
            ---

            # Article
        ''').strip()

        for article in already_outdated_translations:
            with open(article, "r", encoding='utf-8') as fd:
                content = fd.read()

            assert content == expected_content.format(outdater.OUTDATED_TRANSLATION_TAG, outdater.OUTDATED_HASH_TAG, commit_hash_1)

        for article in outdated_translations:
            with open(article, "r", encoding='utf-8') as fd:
                content = fd.read()

            assert content == expected_content.format(outdater.OUTDATED_TRANSLATION_TAG, outdater.OUTDATED_HASH_TAG, commit_hash_2)

        for article in utils.take(article_paths, "en.md"):
            with open(article, "r", encoding='utf-8') as fd:
                content = fd.read()

            assert content == '# Article\n\nThis is an article in English.'

        log = git_utils.git("--no-pager", "log", "--pretty=oneline").splitlines()

        assert len(log) == 4

    def test__full_autofix_flow_with_changed_root(self, root):
        utils.set_up_dummy_repo()
        article_paths = [
            'wiki/Article/en.md',
            'wiki/Article/fr.md',
            'wiki/Article/pt-br.md',
            'wiki/Article/zh-tw.md',
            'wiki/Category1/Article/en.md',
            'wiki/Category1/Article/fr.md',
            'wiki/Category1/Article/pt-br.md',
            'wiki/Category1/Article/zh-tw.md',
            'wiki/Category1/Category2/Article/en.md',
            'wiki/Category1/Category2/Article/fr.md',
            'wiki/Category1/Category2/Article/pt-br.md',
            'wiki/Category1/Category2/Article/zh-tw.md',
            'wiki/Category1/Category2/Category3/Article/en.md',
            'wiki/Category1/Category2/Category3/Article/fr.md',
            'wiki/Category1/Category2/Category3/Article/pt-br.md',
            'wiki/Category1/Category2/Category3/Article/zh-tw.md',
        ]

        utils.create_files(root, *((path, '# Article') for path in article_paths))
        utils.stage_all_and_commit("add articles")
        commit_hash_1 = utils.get_last_commit_hash()

        already_outdated_translations = utils.take(article_paths, "zh-tw.md")
        outdater.outdate_translations(*already_outdated_translations, outdated_hash=commit_hash_1)
        utils.stage_all_and_commit("outdate chinese translations")

        utils.create_files(root, *(
            (article_path, '# Article\n\nThis is an article in English.') for article_path in
            utils.take(article_paths, "en.md")
        ))
        utils.stage_all_and_commit("modify english articles")
        commit_hash_2 = utils.get_last_commit_hash()

        cd = file_utils.ChangeDirectory("wiki")
        exit_code = outdater.main("--root", "..", "--base-commit", commit_hash_2, f"{outdater.AUTOFIX_FLAG}")
        del cd

        assert exit_code == 0

        outdated_translations = utils.get_changed_files()

        non_chinese_translations = utils.remove(article_paths, "en.md", "zh-tw.md")

        assert multiset(outdated_translations) == multiset(non_chinese_translations)

        expected_content = textwrap.dedent('''
            ---
            {}: true
            {}: {}
            ---

            # Article
        ''').strip()

        for article in already_outdated_translations:
            with open(article, "r", encoding='utf-8') as fd:
                content = fd.read()

            assert content == expected_content.format(outdater.OUTDATED_TRANSLATION_TAG, outdater.OUTDATED_HASH_TAG, commit_hash_1)

        for article in outdated_translations:
            with open(article, "r", encoding='utf-8') as fd:
                content = fd.read()

            assert content == expected_content.format(outdater.OUTDATED_TRANSLATION_TAG, outdater.OUTDATED_HASH_TAG, commit_hash_2)

        for article in utils.take(article_paths, "en.md"):
            with open(article, "r", encoding='utf-8') as fd:
                content = fd.read()

            assert content == '# Article\n\nThis is an article in English.'

        log = git_utils.git("--no-pager", "log", "--pretty=oneline").splitlines()

        assert len(log) == 3

    def test__full_autocommit_flow_with_changed_root(self, root):
        utils.set_up_dummy_repo()
        article_paths = [
            'wiki/Article/en.md',
            'wiki/Article/fr.md',
            'wiki/Article/pt-br.md',
            'wiki/Article/zh-tw.md',
            'wiki/Category1/Article/en.md',
            'wiki/Category1/Article/fr.md',
            'wiki/Category1/Article/pt-br.md',
            'wiki/Category1/Article/zh-tw.md',
            'wiki/Category1/Category2/Article/en.md',
            'wiki/Category1/Category2/Article/fr.md',
            'wiki/Category1/Category2/Article/pt-br.md',
            'wiki/Category1/Category2/Article/zh-tw.md',
            'wiki/Category1/Category2/Category3/Article/en.md',
            'wiki/Category1/Category2/Category3/Article/fr.md',
            'wiki/Category1/Category2/Category3/Article/pt-br.md',
            'wiki/Category1/Category2/Category3/Article/zh-tw.md',
        ]

        utils.create_files(root, *((path, '# Article') for path in article_paths))
        utils.stage_all_and_commit("add articles")
        commit_hash_1 = utils.get_last_commit_hash()

        already_outdated_translations = utils.take(article_paths, "zh-tw.md")
        outdater.outdate_translations(*already_outdated_translations, outdated_hash=commit_hash_1)
        utils.stage_all_and_commit("outdate chinese translations")

        utils.create_files(root, *(
            (article_path, '# Article\n\nThis is an article in English.') for article_path in
            utils.take(article_paths, "en.md")
        ))
        utils.stage_all_and_commit("modify english articles")
        commit_hash_2 = utils.get_last_commit_hash()

        cd = file_utils.ChangeDirectory("wiki")
        exit_code = outdater.main("--root", "..", "--base-commit", commit_hash_2, f"{outdater.AUTOFIX_FLAG}", f"{outdater.AUTOCOMMIT_FLAG}")
        del cd

        assert exit_code == 0

        commit_hash_3 = utils.get_last_commit_hash()

        assert commit_hash_3 != commit_hash_2

        assert utils.get_changed_files() == []

        outdated_translations = outdater.list_modified_translations(commit_hash_3)

        non_chinese_translations = utils.remove(article_paths, "en.md", "zh-tw.md")

        assert multiset(outdated_translations) == multiset(non_chinese_translations)

        expected_content = textwrap.dedent('''
            ---
            {}: true
            {}: {}
            ---

            # Article
        ''').strip()

        for article in already_outdated_translations:
            with open(article, "r", encoding='utf-8') as fd:
                content = fd.read()

            assert content == expected_content.format(outdater.OUTDATED_TRANSLATION_TAG, outdater.OUTDATED_HASH_TAG, commit_hash_1)

        for article in outdated_translations:
            with open(article, "r", encoding='utf-8') as fd:
                content = fd.read()

            assert content == expected_content.format(outdater.OUTDATED_TRANSLATION_TAG, outdater.OUTDATED_HASH_TAG, commit_hash_2)

        for article in utils.take(article_paths, "en.md"):
            with open(article, "r", encoding='utf-8') as fd:
                content = fd.read()

            assert content == '# Article\n\nThis is an article in English.'

        log = git_utils.git("--no-pager", "log", "--pretty=oneline").splitlines()

        assert len(log) == 4



import pathlib
import textwrap

import pytest

import tests.conftest
import tests.utils as utils

from wikitools import article_parser, link_checker, link_parser, redirect_parser, errors as error_types, reference_parser


def dummy_article(path):
    return article_parser.Article(
        pathlib.Path(path), lines={}, references={}, identifiers={}, front_matter={}
    )


class TestArticleLinks:
    @pytest.mark.parametrize(
        "payload",
        [
            {"case_sensitive": False, "capitalisation_correct": False, "should_error": False},
            {"case_sensitive": False, "capitalisation_correct": True, "should_error": False},
            {"case_sensitive": True, "capitalisation_correct": False, "should_error": True},
            {"case_sensitive": True, "capitalisation_correct": True, "should_error": False},
        ]
    )
    def test__valid_absolute_article_link(self, root, payload):
        utils.create_files(
            root,
            ('wiki/First_article/en.md', '# First article')
        )

        link = link_parser.find_link('Check the [first article](/wiki/{}).'
            .format("First_article" if payload["capitalisation_correct"] else "First_Article"))
        assert link
        error = link_checker.check_link(
            article=dummy_article('does/not/matter'),
            link=link, redirects={}, references={}, all_articles={},
            case_sensitive=payload["case_sensitive"]
        )
        if payload["should_error"]:
            assert isinstance(error, error_types.LinkNotFoundError)
        else:
            assert error is None

    @pytest.mark.parametrize(
        "payload",
        [
            {"case_sensitive": False, "capitalisation_correct": False, "should_error": False},
            {"case_sensitive": False, "capitalisation_correct": True, "should_error": False},
            {"case_sensitive": True, "capitalisation_correct": False, "should_error": True},
            {"case_sensitive": True, "capitalisation_correct": True, "should_error": False},
        ]
    )
    def test__valid_absolute_image_link(self, root, payload):
        utils.create_files(
            root,
            ('wiki/First_article/img/cat.png', '')
        )

        link = link_parser.find_link('Check the [first article](/wiki/{}/img/cat.png).'
            .format("First_article" if payload["capitalisation_correct"] else "First_Article"))
        assert link
        error = link_checker.check_link(
            article=dummy_article('does/not/matter'),
            link=link, redirects={}, references={}, all_articles={},
            case_sensitive=payload["case_sensitive"]
        )
        if payload["should_error"]:
            assert isinstance(error, error_types.LinkNotFoundError)
        else:
            assert error is None

    @pytest.mark.parametrize(
        "payload",
        [
            {"case_sensitive": False, "capitalisation_correct": False, "should_error": False},
            {"case_sensitive": False, "capitalisation_correct": True, "should_error": False},
            {"case_sensitive": True, "capitalisation_correct": False, "should_error": True},
            {"case_sensitive": True, "capitalisation_correct": True, "should_error": False},
        ]
    )
    def test__valid_absolute_news_link(self, root, payload):
        utils.create_files(
            root,
            ('news/2023/2023-01-01-first-article.md', '# This is a news post')
        )

        link = link_parser.find_link('Check the [first article](https://osu.ppy.sh/home/news/2023-01-01-{}).'
            .format("first-article" if payload["capitalisation_correct"] else "First-Article"))
        assert link
        error = link_checker.check_link(
            article=dummy_article('does/not/matter'),
            link=link, redirects={}, references={}, all_articles={},
            case_sensitive=payload["case_sensitive"]
        )
        if payload["should_error"]:
            assert isinstance(error, error_types.LinkNotFoundError)
        else:
            assert error is None

    def test__invalid_absolute_link(self, root):
        utils.create_files(
            root,
            ('wiki/Another_article/en.md', '# Another article')
        )

        for line in (
            'This link is [broken](/wiki/Another_article/Some_nonsense).',
            'This link is [broken](/wiki/A_random_directory), too.'
        ):
            link = link_parser.find_link(line)
            assert link
            error = link_checker.check_link(
                article=dummy_article('does/not/matter'), link=link, redirects={}, references={}, all_articles={}
            )
            assert isinstance(error, error_types.LinkNotFoundError)

    def test__invalid_link_with_filename(self, root):
        utils.create_files(
            root,
            ('wiki/Another_article/en.md', '# Another article')
        )

        link = link_parser.find_link("This link [shouldn't include the filename](/wiki/Another_article/en.md).")
        assert link
        error = link_checker.check_link(
            article=dummy_article('does/not/matter'), link=link, redirects={}, references={}, all_articles={}
        )
        assert isinstance(error, error_types.MalformedLinkError)

    def test__valid_reference(self, root):
        utils.create_files(
            root,
            ('wiki/My_article/en.md', '# My article')
        )

        link = link_parser.find_link('This link is [working][article_ref].')
        assert link
        references = reference_parser.extract_all('[article_ref]: /wiki/My_article "Something something"')
        error = link_checker.check_link(
            article=dummy_article('does/not/matter'), link=link, redirects={}, references=references, all_articles={}
        )
        assert error is None

    def test__invalid_reference(self, root):
        utils.create_files(
            root,
            ('wiki/Obscure_article/en.md', '# First article')
        )

        link = link_parser.find_link('This link is [not working][article_ref].')
        assert link
        references = reference_parser.extract_all('[other_ref]: /wiki/Obscure_article "Something something"')
        error = link_checker.check_link(
            article=dummy_article('does/not/matter'), link=link, redirects={}, references=references, all_articles={}
        )
        assert isinstance(error, error_types.MissingReferenceError)
        assert error.link.raw_location == 'article_ref'

    def test__valid_relative_link(self, root):
        utils.create_files(
            root,
            ('wiki/Batteries/en.md', '# Batteries'),
            ('wiki/Batteries/Included/en.md', '# Included'),
            ('wiki/Batteries/Included/And_even_more/en.md', '# And even more!')
        )

        for line in (
            '[Alkaline FTW](Included).',
            '[Alkaline FTW](./Included).',
            '[Alkaline FTW](../Batteries/Included).',  # I hope we will never see this, but let's be ready
            '[Alkaline FTW](Included/And_even_more).'
        ):
            link = link_parser.find_link(line)
            assert link
            error = link_checker.check_link(
                article=dummy_article('wiki/Batteries/en.md'), link=link, redirects={}, references={}, all_articles={}
            )
            assert error is None

    def test__invalid_relative_link(self, root):
        utils.create_files(
            root,
            ('wiki/Existing_article/en.md', '# Existing article')
        )

        link = link_parser.find_link('This link [does not work](Broken_link).')
        assert link
        error = link_checker.check_link(
            article=dummy_article('wiki/Existing_article/en.md'), link=link, redirects={}, references={}, all_articles={}
        )
        assert isinstance(error, error_types.LinkNotFoundError)


class TestImageLinks:
    def test__valid_absolute_link(self, root):
        utils.create_files(
            root,
            ('wiki/Article/en.md', '# Article'),
            ('wiki/img/battery.png', '')
        )

        link = link_parser.find_link('Check this ![out](/wiki/img/battery.png).')
        assert link
        error = link_checker.check_link(
            article=dummy_article('does/not/matter'), link=link, redirects={}, references={}, all_articles={}
        )
        assert error is None

    def test__invalid_absolute_link(self, root):
        utils.create_files(
            root,
            ('wiki/New_article/en.md', '# New article'),
            ('wiki/img/battery.png', '')
        )

        link = link_parser.find_link('Do not check this ![out](/wiki/img/nonsense.png).')
        assert link
        error = link_checker.check_link(
            article=dummy_article('does/not/matter'), link=link, redirects={}, references={}, all_articles={}
        )
        assert isinstance(error, error_types.LinkNotFoundError)

    def test__valid_relative_link(self, root):
        utils.create_files(
            root,
            ('wiki/Beatmap/en.md', '# Beatmap'),
            ('wiki/Beatmap/img/beatmap.png', '')
        )

        link = link_parser.find_link('Behold, the beatmap ![beatmap](img/beatmap.png "Wow!").')
        assert link
        error = link_checker.check_link(
            article=dummy_article('wiki/Beatmap/en.md'), link=link, redirects={}, references={}, all_articles={}
        )
        assert error is None

    # this is wrong as per ASC but still semantically correct
    def test__valid_relative_link_in_article_directory(self, root):
        utils.create_files(
            root,
            ('wiki/Beatmap/en.md', '# Beatmap'),
            ('wiki/Beatmap/beatmap.png', '')
        )

        link = link_parser.find_link('Behold, the relative image of a ![beatmap](beatmap.png "Wow!").')
        assert link
        error = link_checker.check_link(
            article=dummy_article('wiki/Beatmap/en.md'), link=link, redirects={}, references={}, all_articles={}
        )
        assert error is None

    def test__invalid_relative_link(self, root):
        utils.create_files(
            root,
            ('wiki/Difficulty/en.md', '# Difficulty'),
            ('wiki/Difficulty/img/difficulty.png', '')
        )

        link = link_parser.find_link('Nothing to see here ![please](img/none.png "disperse").')
        assert link
        error = link_checker.check_link(
            article=dummy_article('wiki/Difficulty/en.md'), link=link, redirects={}, references={}, all_articles={}
        )
        assert isinstance(error, error_types.LinkNotFoundError)

    def test__invalid_reference_link(self, root):
        utils.create_files(
            root,
            ('wiki/OWC_2030/en.md', '# OWC 2030'),
            ('wiki/img/dummy.png', '')
        )

        references = reference_parser.extract_all('[flag_XX]: /wiki/shared/img/XX.gif')
        link = link_parser.find_link('![][flag_XX] "The XXth Country"')
        assert link
        error = link_checker.check_link(
            article=dummy_article('wiki/OWC_2030/en.md'), link=link, redirects={}, references=references, all_articles={}
        )
        assert isinstance(error, error_types.LinkNotFoundError)
        assert isinstance(error.link, link_parser.Link)
        assert error.link.is_reference
        assert error.link.raw_location == 'flag_XX'
        assert error.reference
        assert error.reference.raw_location == '/wiki/shared/img/XX.gif'
        assert error.reference.lineno == 1


class TestRedirectedLinks:
    def test__valid_link(self, root):
        utils.create_files(
            root,
            ('wiki/redirect.yaml', '"old_link": "New_article"'),
            ('wiki/New_article/en.md', '# New article'),
        )

        redirects = redirect_parser.load_redirects('wiki/redirect.yaml')
        link = link_parser.find_link('Please read the [old article](/wiki/Old_LiNK).')
        assert link
        error = link_checker.check_link(
            article=dummy_article('does/not/matter'), link=link, redirects=redirects, references={}, all_articles={}
        )
        assert error is None

    def test__invalid_link(self, root):
        utils.create_files(
            root,
            (
                'wiki/redirect.yaml', textwrap.dedent('''
                    # junk comment to fill the lines
                    "old_link": "Wrong_redirect"
                ''').strip()
            ),
            ('wiki/New_article/en.md', '# New article'),
        )

        redirects = redirect_parser.load_redirects('wiki/redirect.yaml')
        link = link_parser.find_link('Please read the [old article](/wiki/Old_link).')
        assert link
        error = link_checker.check_link(
            article=dummy_article('does/not/matter'), link=link, redirects=redirects, references={}, all_articles={}
        )
        assert isinstance(error, error_types.BrokenRedirectError)
        assert error.redirect_lineno == 2
        assert error.resolved_location == 'Old_link'
        assert error.redirect_destination == 'Wrong_redirect'


class TestNewspostLinks:
    def test__valid_newspost_link(self, root):
        utils.create_files(
            root,
            (
                'news/2007/2007-01-01-newspost.md', textwrap.dedent('''
                    ---
                    layout: post
                    title: News!!!
                    date: 2007-01-01 12:00:00 +0000
                    ---

                    Today we have big news!!!!
                ''').strip()
            ),
        )
        link = link_parser.find_link('Please read the [latest news post](https://osu.ppy.sh/home/news/2007-01-01-newspost).')
        assert link
        error = link_checker.check_link(
            article=dummy_article('does/not/matter'), link=link, redirects={}, references={}, all_articles={}
        )
        assert error is None

    def test__invalid_newspost_link(self, root):
        utils.create_files(
            root,
            (
                'news/2007/2007-01-01-newspost.md', textwrap.dedent('''
                    ---
                    layout: post
                    title: News!!!
                    date: 2007-01-01 12:00:00 +0000
                    ---

                    Today we have big news!!!!
                ''').strip()
            ),
        )
        link = link_parser.find_link('Please read the [latest news post](https://osu.ppy.sh/home/news/2007-01-01-not-a-newspost).')
        assert link
        error = link_checker.check_link(
            article=dummy_article('does/not/matter'), link=link, redirects={}, references={}, all_articles={}
        )
        assert error
        assert isinstance(error, error_types.LinkNotFoundError)


class TestNewspostSectionLinks:
    @pytest.mark.parametrize(
        "payload",
        [
            {"case_sensitive": False, "capitalisation_correct": False, "should_error": False},
            {"case_sensitive": False, "capitalisation_correct": True, "should_error": False},
            {"case_sensitive": True, "capitalisation_correct": False, "should_error": True},
            {"case_sensitive": True, "capitalisation_correct": True, "should_error": False},
        ]
    )
    def test__section_link_casing(self, root, payload):
        utils.create_files(
            root,
            ('wiki/First_article/en.md', '## First article')
        )

        link = link_parser.find_link('Check the [first article](/wiki/{}#first-article).'
            .format("First_article" if payload["capitalisation_correct"] else "First_Article"))
        assert link
        error = link_checker.check_link(
            article=dummy_article('does/not/matter'),
            link=link, redirects={}, references={}, all_articles={},
            case_sensitive=payload["case_sensitive"]
        )
        if payload["should_error"]:
            assert isinstance(error, error_types.LinkNotFoundError)
        else:
            assert error is None

    def test__valid_newspost_section_link_within_post(self, root):
        utils.create_files(
            root,
            (
                'news/2007/2007-01-01-newspost.md', textwrap.dedent('''
                    ---
                    layout: post
                    title: News!!!
                    date: 2007-01-01 12:00:00 +0000
                    ---

                    Today we have big news!!!!

                    Skip to [the news](#the-news)!

                    ## The news
                ''').strip()
            ),
        )
        article = article_parser.parse("news/2007/2007-01-01-newspost.md")
        assert article.identifiers == {'the-news': 11}
        all_articles = {article.path: article}

        assert link_checker.check_article(article=article, redirects={}, all_articles={}) == {}

    def test__valid_newspost_section_link(self, root):
        utils.create_files(
            root,
            (
                'news/2007/2007-01-01-newspost.md', textwrap.dedent('''
                    ---
                    layout: post
                    title: News!!!
                    date: 2007-01-01 12:00:00 +0000
                    ---

                    Today we have big news!!!!

                    ## The news
                ''').strip()
            ),
        )
        article = article_parser.parse("news/2007/2007-01-01-newspost.md")
        assert article.identifiers == {'the-news': 9}
        all_articles = {article.path: article}

        link = link_parser.find_link('Please read the [latest news post](https://osu.ppy.sh/home/news/2007-01-01-newspost#the-news).')
        assert link
        error = link_checker.check_link(
            article=dummy_article('does/not/matter'), link=link, redirects={}, references={}, all_articles=all_articles
        )
        assert error is None

    def test__invalid_newspost_section_link(self, root):
        utils.create_files(
            root,
            (
                'news/2007/2007-01-01-newspost.md', textwrap.dedent('''
                    ---
                    layout: post
                    title: News!!!
                    date: 2007-01-01 12:00:00 +0000
                    ---

                    Today we have big news!!!!

                    ## The news
                ''').strip()
            ),
        )
        article = article_parser.parse("news/2007/2007-01-01-newspost.md")
        assert article.identifiers == {'the-news': 9}
        all_articles = {article.path: article}

        link = link_parser.find_link('Please read the [latest news post](https://osu.ppy.sh/home/news/2007-01-01-newspost#the-fake-news).')
        assert link
        error = link_checker.check_link(
            article=dummy_article('does/not/matter'), link=link, redirects={}, references={}, all_articles=all_articles
        )
        assert error
        assert isinstance(error, error_types.MissingIdentifierError)
        assert error.link == link
        assert error.path == "news/2007/2007-01-01-newspost.md"
        assert error.identifier == "the-fake-news"
        assert not error.no_translation_available
        assert not error.translation_outdated


class TestExternalLinks:
    def test__all_external_links__valid(self):
        for line in (
            'Check the [example](https://example.com "Example").',
            'Check [this Wikipedia article](https://en.wikipedia.org/wiki/MediaWiki "MediaWiki on Wikipedia").',
            'Contact [accounts@example.com](mailto:accounts@example.com).',
            'Look, [the web chat](irc://cho.ppy.sh)!',
            'I am [not even trying](htttttttttttttttttps://example.com).',
        ):
            link = link_parser.find_link(line)
            assert link
            error = link_checker.check_link(
                article=dummy_article('does/not/matter'), link=link, redirects={}, references={}, all_articles={}
            )
            assert error is None

    def test__all_external_reference_links__valid(self):
        for line, reference in (
            ('Check the [example][example].', '[example]: https://example.com "Example"'),
            ('Contact [accounts@example.com][email].', '[email]: mailto:accounts@example.com'),
            ('Look, [the web chat][irc]!', '[irc]: irc://cho.ppy.sh'),
            ('I am [not even trying][aaa].', '[aaa]: htttttttttttttttttps://example.com'),
        ):
            link = link_parser.find_link(line)
            assert link
            references = reference_parser.extract_all(reference)
            error = link_checker.check_link(
                article=dummy_article('does/not/matter'), link=link, redirects={}, references=references, all_articles={}
            )
            assert error is None


class TestGitHubLinks:
    @pytest.mark.parametrize(
        "payload",
        [
            {"string": "This link is [broken](https://github.com/ppy/osu-wiki/blob/master/wiki/Another_article/Some_nonsense/en.md).", "resolved_location": "wiki/Another_article/Some_nonsense/en.md"},
            {"string": "This link is [broken](https://github.com/ppy/osu-wiki/tree/master/wiki/A_random_directory), too.", "resolved_location": "wiki/A_random_directory"},
            {"string": "[casing matters](https://github.com/ppy/osu-wiki/tree/master/wiki/Another_Article), too.", "resolved_location": "wiki/Another_Article"},
        ]
    )
    def test__github_link_invalid(self, root, payload):
        utils.create_files(
            root,
            ('wiki/Another_article/en.md', '# Another article')
        )

        link = link_parser.find_link(payload["string"])
        assert link
        error = link_checker.check_link(
            article=dummy_article('does/not/matter'), link=link, redirects={}, references={}, all_articles={}, case_sensitive=False
        )
        assert isinstance(error, error_types.LinkNotFoundError)
        assert error.resolved_location == payload["resolved_location"]
        error = link_checker.check_link(
            article=dummy_article('does/not/matter'), link=link, redirects={}, references={}, all_articles={}, case_sensitive=True
        )
        assert isinstance(error, error_types.LinkNotFoundError)
        assert error.resolved_location == payload["resolved_location"]

    @pytest.mark.parametrize(
        "payload",
        [
            {"string": "This link is [ok](https://github.com/ppy/osu-wiki/blob/master/wiki/Another_article/en.md).", "resolved_location": "wiki/Another_article/en.md"},
            {"string": "This link is [also ok](https://github.com/ppy/osu-wiki/tree/master/wiki/Another_article).", "resolved_location": "wiki/Another_article"},
            {"string": "This link is [very ok](https://github.com/ppy/osu-wiki/tree/master/news).", "resolved_location": "news"},
            {"string": "This link is [another ok link](https://github.com/ppy/osu-wiki/tree/master/README.md).", "resolved_location": "README.md"},
        ]
    )
    def test__github_link_valid(self, root, payload):
        utils.create_files(
            root,
            ('wiki/Another_article/en.md', '# Another article'),
            ('news/2023/newspost.md', '# news post'),
            ('README.md', '# Please read me')
        )

        link = link_parser.find_link(payload["string"])
        assert link
        error = link_checker.check_link(
            article=dummy_article('does/not/matter'), link=link, redirects={}, references={}, all_articles={}
        )
        assert error is None

    @pytest.mark.parametrize(
        "payload",
        [
            {"string": "Please read the [article](https://github.com/ppy/osu-wiki/blob/master/wiki/New_article/en.md#some-nonexistent-heading).", "resolved_location": "wiki/New_article/en.md"},
            {"string": "Please read the [readme](https://github.com/ppy/osu-wiki/blob/master/README.md#some-nonexistent-heading).", "resolved_location": "README.md"},
        ]
    )
    def test__github_link_invalid__missing_heading(self, root, payload):
        utils.create_files(
            root,
            ('wiki/New_article/en.md', '# New article'),
            ('README.md', '# Please read me')
        )
        new_article = dummy_article('wiki/New_article/en.md')
        all_articles = {new_article.path: new_article}

        link = link_parser.find_link(payload["string"])
        assert link
        error = link_checker.check_link(
            article=dummy_article('wiki/Other_article/en.md'), link=link, redirects={}, references={}, all_articles=all_articles
        )
        assert isinstance(error, error_types.MissingIdentifierError)
        assert error.identifier == 'some-nonexistent-heading'
        assert error.path == payload["resolved_location"]
        assert not error.no_translation_available
        assert not error.translation_outdated


class TestMalformedLink:
    def test__missing_scheme(self):
        link = link_parser.find_link('Forgot to add a [scheme](//example.com)',)
        assert link
        error = link_checker.check_link(
            article=dummy_article('does/not/matter'), link=link, redirects={}, references={}, all_articles={}
        )
        assert isinstance(error, error_types.MalformedLinkError)
        assert error.link.raw_location == '//example.com'


class TestSectionLinks:
    def test__valid_section_link_within_article(self, root):
        utils.create_files(
            root,
            (
                'wiki/New_article/en.md',
                textwrap.dedent('''
                    # New article

                    ## Some real heading

                    Some real though from a real person.

                    check out [this section](#some-real-heading)
                ''').strip()
            ),
        )
        article = article_parser.parse("wiki/New_article/en.md")
        assert article.identifiers == {'some-real-heading': 3}
        all_articles = {article.path: article}

        assert link_checker.check_article(article=article, redirects={}, all_articles={}) == {}

    def test__valid_absolute_link(self, root):
        utils.create_files(
            root,
            (
                'wiki/New_article/en.md',
                textwrap.dedent('''
                    # New article

                    ## Some real heading

                    Some real though from a real person.
                ''').strip()
            )
        )
        new_article = article_parser.parse('wiki/New_article/en.md')
        assert new_article.identifiers == {'some-real-heading': 3}
        all_articles = {new_article.path: new_article}

        link = link_parser.find_link('Please read the [article](/wiki/New_article#some-real-heading).')
        assert link
        error = link_checker.check_link(
            article=dummy_article('does/not/matter'), link=link, redirects={}, references={}, all_articles=all_articles
        )
        assert error is None

    def test__valid_absolute_link__translation(self, root):
        utils.create_files(
            root,
            ('wiki/New_article/en.md', '# New article'),
            (
                'wiki/New_article/ru.md',
                textwrap.dedent(u'''
                    # New article

                    ## Заголовок (translated)
                ''')
            )
        )

        all_articles = {
            path: article_parser.parse(path)
            for path in ('wiki/New_article/en.md', 'wiki/New_article/ru.md')
        }

        link = link_parser.find_link('См. [другую статью](/wiki/New_article#заголовок-(translated)).')
        assert link
        error = link_checker.check_link(
            article=dummy_article('wiki/Some_other_article/ru.md'), link=link, redirects={}, references={}, all_articles=all_articles
        )
        assert error is None

    def test__invalid_absolute_link__missing_heading(self, root):
        utils.create_files(
            root,
            ('wiki/New_article/en.md', '# New article'),
        )
        new_article = dummy_article('wiki/New_article/en.md')
        all_articles = {new_article.path: new_article}

        link = link_parser.find_link('Please read the [article](/wiki/New_article#some-nonexistent-heading).')
        assert link
        error = link_checker.check_link(
            article=dummy_article('wiki/Other_article/en.md'), link=link, redirects={}, references={}, all_articles=all_articles
        )
        assert isinstance(error, error_types.MissingIdentifierError)
        assert error.identifier == 'some-nonexistent-heading'
        assert error.path == 'wiki/New_article/en.md'
        assert not error.no_translation_available
        assert not error.translation_outdated

    def test__invalid_absolute_link__missing_translation(self, root):
        utils.create_files(
            root,
            (
                'wiki/New_article/en.md',
                textwrap.dedent('''
                    # New article

                    ## Self-check

                    This line exists.
                ''').strip()
            )
        )
        new_article = article_parser.parse('wiki/New_article/en.md')
        all_articles = {new_article.path: new_article}

        # pretend we're linking from a French page
        link = link_parser.find_link("Merci de lire l'[article](/wiki/New_article#auto-contrôle).")
        assert link
        error = link_checker.check_link(
            article=dummy_article('wiki/Some_other_article/fr.md'), link=link, redirects={}, references={}, all_articles=all_articles
        )
        assert isinstance(error, error_types.MissingIdentifierError)
        assert error.identifier == 'auto-contrôle'
        assert error.path == 'wiki/New_article/en.md'
        assert error.no_translation_available
        assert not error.translation_outdated

    def test__valid_relative_link(self, root):
        utils.create_files(
            root,
            ('wiki/New_article/en.md', '# New article'),
            (
                'wiki/New_article/Included_article/en.md',
                textwrap.dedent('''
                    # Included article

                    ## Subheading

                    This line exists.
                ''').strip()
            )
        )
        all_articles = {
            path: article_parser.parse(path)
            for path in ('wiki/New_article/en.md', 'wiki/New_article/Included_article/en.md')
        }

        link = link_parser.find_link("Please follow the [included article](Included_article#subheading).")
        assert link
        error = link_checker.check_link(
            article=dummy_article('wiki/New_article/en.md'), link=link, redirects={}, references={}, all_articles=all_articles
        )
        assert error is None

    def test__invalid_relative_link(self, root):
        utils.create_files(
            root,
            ('wiki/New_article/en.md', '# New article'),
            (
                'wiki/New_article/Included_article/en.md',
                textwrap.dedent('''
                    # Included article

                    ## Subheading

                    This line exists.
                ''').strip()
            )
        )
        all_articles = {
            path: article_parser.parse(path)
            for path in ('wiki/New_article/en.md', 'wiki/New_article/Included_article/en.md')
        }

        link = link_parser.find_link("Please follow the [included article](Included_article#wrong-subheading).")
        assert link
        error = link_checker.check_link(
            article=dummy_article('wiki/New_article/en.md'), link=link, redirects={}, references={}, all_articles=all_articles
        )
        assert isinstance(error, error_types.MissingIdentifierError)
        assert error.identifier == 'wrong-subheading'
        assert error.path == 'wiki/New_article/Included_article/en.md'

    def test__valid_redirect(self, root):
        utils.create_files(
            root,
            ('wiki/redirect.yaml', '"old_location": "Target_article"'),
            ('wiki/New_article/en.md', '# New article'),
            (
                'wiki/Target_article/en.md',
                textwrap.dedent('''
                    # Included article

                    ## Subheading

                    This line exists.
                ''').strip()
            )
        )
        all_articles = {
            path: article_parser.parse(path)
            for path in ('wiki/New_article/en.md', 'wiki/Target_article/en.md')
        }
        redirects = redirect_parser.load_redirects('wiki/redirect.yaml')

        link = link_parser.find_link("Please follow the [target article](/wiki/Old_location#subheading).")
        assert link
        error = link_checker.check_link(
            article=dummy_article('wiki/New_article/en.md'), link=link, redirects=redirects, references={}, all_articles=all_articles
        )
        assert error is None

    def test__invalid_redirect(self, root):
        utils.create_files(
            root,
            ('wiki/redirect.yaml', '"old_location": "Target_article"'),
            ('wiki/New_article/en.md', '# New article'),
            (
                'wiki/Target_article/en.md',
                textwrap.dedent('''
                    # Included article

                    ## Subheading

                    This line exists.
                ''').strip()
            )
        )
        all_articles = {
            path: article_parser.parse(path)
            for path in ('wiki/New_article/en.md', 'wiki/Target_article/en.md')
        }
        redirects = redirect_parser.load_redirects('wiki/redirect.yaml')

        link = link_parser.find_link("Please follow the [target article](/wiki/Old_location#totally-wrong-heading).")
        assert link
        error = link_checker.check_link(
            article=dummy_article('wiki/New_article/en.md'), link=link, redirects=redirects, references={}, all_articles=all_articles
        )
        assert isinstance(error, error_types.MissingIdentifierError)
        assert error.identifier == 'totally-wrong-heading'
        assert error.path == 'wiki/Target_article/en.md'


class TestArticleChecker:
    def test__check_article__clean(self, root):
        utils.create_files(
            root,
            (
                'wiki/Article/en.md',
                textwrap.dedent('''
                    # An article

                    It's a [very](Subarticle) [well](/wiki/Brticle) written ![line](img/line.png). [[1]][r]

                    This line, however, uses a [redirect](/wiki/Old_link).

                    ## References

                    1. yes

                    [r]: #references
                ''').strip()
            ),
            ('wiki/Article/Subarticle/en.md', ''),
            ('wiki/Brticle/en.md', ''),
            ('wiki/Article/img/line.png', ''),
            ('wiki/redirect.yaml', '"old_link": "Article/Subarticle"')
        )

        redirects = redirect_parser.load_redirects('wiki/redirect.yaml')
        article = article_parser.parse('wiki/Article/en.md')
        assert link_checker.check_article(article, redirects, all_articles={}) == {}

    def test__check_article__bad(self, root):
        utils.create_files(
            root,
            (
                'wiki/Article/en.md',
                textwrap.dedent('''
                    # An article

                    [Some](/wiki/Valid_link) of the lines contain valid links and images ![yep](img/yep.png).

                    ## Disaster

                    [What?](/wiki/Broken_link)

                    - No, really, [what is this](Bad_relative_link)?
                    - Did you even [check your links](/wiki/Broken_redirect)? [Are you sure?][sure_ref]
                    - **No formatting *[at all][at_all_ref]?!*
                    - Subpar ![a dismissive picture](img/you_tried.jpeg)

                    ## References

                    [sure_ref]: /wiki/Article
                ''').strip()
            ),
            ('wiki/Article/img/yep.png', ''),
            ('wiki/Valid_link/en.md', ''),
            ('wiki/redirect.yaml', '"broken_redirect": "Another_missing_article"')
        )

        redirects = redirect_parser.load_redirects('wiki/redirect.yaml')
        article = article_parser.parse('wiki/Article/en.md')
        errors = link_checker.check_article(article, redirects, all_articles={})
        assert sum(len(ee) for ee in errors.values()) == 5

        flattened_errors = []
        for lineno, rr in sorted(errors.items()):
            for result in rr:
                flattened_errors.append((lineno, result))

        assert set(r.link.raw_location for (_, r) in flattened_errors) == {
            '/wiki/Broken_link',
            'Bad_relative_link',
            '/wiki/Broken_redirect',
            'at_all_ref',
            'img/you_tried.jpeg'
        }

        broken_link_error = flattened_errors[0][1]
        broken_link = flattened_errors[0][1].link
        assert isinstance(broken_link_error, error_types.LinkNotFoundError)
        assert broken_link_error.resolved_location == 'wiki/Broken_link'
        assert (flattened_errors[0][0], broken_link.start) == (7, 0)

        broken_rel_link_error = flattened_errors[1][1]
        broken_rel_link = flattened_errors[1][1].link
        assert isinstance(broken_rel_link_error, error_types.LinkNotFoundError)
        assert broken_rel_link_error.resolved_location == 'wiki/Article/Bad_relative_link'
        assert (flattened_errors[1][0], broken_rel_link.start) == (9, 14)

        broken_redirect_error = flattened_errors[2][1]
        broken_redirect = flattened_errors[2][1].link
        assert isinstance(broken_redirect_error, error_types.BrokenRedirectError)
        assert broken_redirect_error.resolved_location == 'Broken_redirect'
        assert (flattened_errors[2][0], broken_redirect.start) == (10, 15)

        broken_redirect_error = flattened_errors[3][1]
        broken_redirect_2 = flattened_errors[3][1].link
        assert isinstance(broken_redirect_error, error_types.MissingReferenceError)
        assert broken_redirect_error.link.raw_location == 'at_all_ref'
        assert (flattened_errors[3][0], broken_redirect_2.start) == (11, 19)

        broken_image_error = flattened_errors[4][1]
        broken_image = flattened_errors[4][1].link
        assert isinstance(broken_image_error, error_types.LinkNotFoundError)
        assert broken_image_error.resolved_location == 'wiki/Article/img/you_tried.jpeg'
        assert (flattened_errors[4][0], broken_image.start) == (12, 10)

        # all lines, even with references, were cached
        assert all(_[0] in article.lines for _ in flattened_errors)
